import traceback

from typing import Dict, List, Optional

from clipped.utils.np import sanitize_dict, sanitize_np_types

from hypertune.logger import logger
from polyaxon.client import RunClient
from polyaxon.schemas import V1Join
from traceml.artifacts import V1ArtifactKind, V1RunArtifact


def get_iteration_definition(
    client: RunClient,
    iteration: int,
    join: V1Join,
    optimization_metric: str,
    name: Optional[str] = None,
):
    def handler():
        runs = (
            client.list(
                query=join.query,
                sort=join.sort,
                limit=join.limit,
                offset=join.offset,
            ).results
            or []
        )
        configs = []
        metrics = []
        run_uuids = []
        for run in runs:
            if optimization_metric in run.outputs:
                run_uuids.append(run.uuid)
                configs.append(run.inputs)
                metrics.append(run.outputs[optimization_metric])

        if configs or metrics or run_uuids:
            artifact_run = V1RunArtifact.construct(
                name=name or "in-iteration-{}".format(iteration),
                kind=V1ArtifactKind.ITERATION,
                summary={
                    "iteration": iteration,
                    "configs": [sanitize_dict(s) for s in configs],
                    "metrics": [sanitize_np_types(s) for s in metrics],
                    "uuid": run_uuids,
                },
                is_input=True,
            )
            client.log_artifact_lineage(artifact_run)

        return run_uuids, configs, metrics

    try:
        return handler()
    except Exception as e:
        exp = "Polyaxon tuner failed fetching iteration definition: {}\n{}".format(
            repr(e), traceback.format_exc()
        )
        client.log_failed(reason="PolyaxonTunerIteration", message=exp)
        logger.warning(e)


def handle_iteration_failure(client: RunClient, exp: Exception):
    exp = "Polyaxon tuner failed creating suggestions: {}\n{}".format(
        repr(exp), traceback.format_exc()
    )
    client.log_failed(reason="PolyaxonTunerSuggestions", message=exp)


def handle_iteration(
    client: RunClient,
    suggestions: List[Dict] = None,
):
    if not suggestions:
        logger.warning("No new suggestions were created")
        return
    try:
        logger.info("Generated new {} suggestions".format(len(suggestions)))
        client.log_outputs(
            suggestions=[sanitize_dict(s) for s in suggestions], async_req=False
        )
        return
    except Exception as e:
        logger.warning(e)
        exp = e

    if exp:
        message = "Polyaxon tuner failed logging iteration definition: {}\n{}".format(
            repr(exp), traceback.format_exc()
        )
        client.log_failed(reason="PolyaxonTunerIteration", message=message)
        raise exp



from typing import Dict, List

from hypertune.search_managers.base import BaseManager
from hypertune.search_managers.bayesian_optimization.optimizer import BOOptimizer
from hypertune.search_managers.random_search.manager import RandomSearchManager
from polyaxon.schemas import V1Bayes, V1RandomSearch


class BayesSearchManager(BaseManager):
    """Bayesian optimization strategy manager for hyperparameter optimization."""

    CONFIG = V1Bayes

    def __init__(self, config):
        super().__init__(config=config)
        self.num_initial_runs = self.config.num_initial_runs
        self.max_iterations = self.config.max_iterations

    def get_suggestions(
        self, configs: List[Dict] = None, metrics: List[float] = None
    ) -> List[Dict]:
        if not configs or not metrics:
            config = V1RandomSearch(
                params=self.config.params,
                num_runs=self.num_initial_runs,
                seed=self.config.seed,
            )
            return RandomSearchManager(config=config).get_suggestions()

        optimizer = BOOptimizer(config=self.config)
        optimizer.add_observations(configs=configs, metrics=metrics)
        suggestion = optimizer.get_suggestion()
        return [suggestion] if suggestion else None



import numpy as np

from datetime import date, datetime

from clipped.compact.pydantic import PYDANTIC_VERSION, ValidationError

from hypertune.matrix import dist
from polyaxon.schemas import (
    V1HpChoice,
    V1HpDateRange,
    V1HpDateTimeRange,
    V1HpGeomSpace,
    V1HpLinSpace,
    V1HpLogNormal,
    V1HpLogSpace,
    V1HpLogUniform,
    V1HpNormal,
    V1HpPChoice,
    V1HpQLogNormal,
    V1HpQLogUniform,
    V1HpQNormal,
    V1HpQUniform,
    V1HpRange,
    V1HpUniform,
    validate_pchoice,
)


def pchoice(values, size=None, rand_generator=None):
    rand_generator = rand_generator or np.random
    keys = [v[0] for v in values]
    dists = [v[1] for v in values]
    validate_pchoice(dists)
    indices = rand_generator.multinomial(1, dists, size=size)
    if size is None:
        return keys[indices.argmax()]
    return [keys[ind.argmax()] for ind in indices]


def space_sample(value, size, rand_generator):
    size = None if size == 1 else size
    rand_generator = rand_generator or np.random
    try:
        return rand_generator.choice(value, size=size)
    except ValueError:
        idx = rand_generator.randint(0, len(value))
        return value[idx]


def space_get_index(array, value):
    try:
        return array.index(value)
    except (ValueError, AttributeError):
        return int(np.where(array == value)[0][0])


def dist_sample(fct, value, size, rand_generator):
    size = None if size == 1 else size
    rand_generator = rand_generator or np.random
    value["size"] = size
    value["rand_generator"] = rand_generator
    return fct(**value)


def get_length(matrix):
    if matrix._IDENTIFIER == V1HpChoice._IDENTIFIER:
        return len(matrix.value)

    if matrix._IDENTIFIER == V1HpPChoice._IDENTIFIER:
        return len(matrix.value)

    if matrix._IDENTIFIER == V1HpDateRange._IDENTIFIER:
        return len(np.arange(**matrix.value.to_dict()))

    if matrix._IDENTIFIER == V1HpDateTimeRange._IDENTIFIER:
        return len(np.arange(**matrix.value.to_dict()))

    if matrix._IDENTIFIER == V1HpRange._IDENTIFIER:
        return len(np.arange(**matrix.value.to_dict()))

    if matrix._IDENTIFIER == V1HpLinSpace._IDENTIFIER:
        return len(np.linspace(**matrix.value.to_dict()))

    if matrix._IDENTIFIER == V1HpLogSpace._IDENTIFIER:
        return len(np.logspace(**matrix.value.to_dict()))

    if matrix._IDENTIFIER == V1HpGeomSpace._IDENTIFIER:
        return len(np.geomspace(**matrix.value.to_dict()))

    if matrix._IDENTIFIER in {
        V1HpUniform._IDENTIFIER,
        V1HpQUniform._IDENTIFIER,
        V1HpLogUniform._IDENTIFIER,
        V1HpQLogUniform._IDENTIFIER,
        V1HpNormal._IDENTIFIER,
        V1HpQNormal._IDENTIFIER,
        V1HpLogNormal._IDENTIFIER,
        V1HpQLogNormal._IDENTIFIER,
    }:
        error_msg = "Distribution should not call `length`"
        if PYDANTIC_VERSION.startswith("2."):
            raise ValidationError.from_exception_data(
                title="Value Error",
                line_errors=[
                    {
                        "type": "value_error",
                        "loc": ("get_length",),
                        # Location tuple indicating where the error occurred
                        "msg": error_msg,
                        "input": matrix.__class__,  # The invalid input that caused the error
                        "ctx": {"error": error_msg},
                    }
                ],
            )
        else:
            raise ValidationError(
                ["Distribution should not call `length`"], matrix.__class__
            )


def get_min(matrix):
    if matrix._IDENTIFIER == V1HpChoice._IDENTIFIER:
        if matrix.is_categorical:
            return None
        return min(to_numpy(matrix))

    if matrix._IDENTIFIER == V1HpPChoice._IDENTIFIER:
        return None

    if matrix._IDENTIFIER in {
        V1HpDateRange._IDENTIFIER,
        V1HpDateTimeRange._IDENTIFIER,
        V1HpRange._IDENTIFIER,
        V1HpLinSpace._IDENTIFIER,
        V1HpLogSpace._IDENTIFIER,
        V1HpGeomSpace._IDENTIFIER,
    }:
        return matrix.value.start

    if matrix._IDENTIFIER == V1HpUniform._IDENTIFIER:
        return matrix.value.low

    if matrix._IDENTIFIER in {
        V1HpQUniform._IDENTIFIER,
        V1HpLogUniform._IDENTIFIER,
        V1HpQLogUniform._IDENTIFIER,
        V1HpNormal._IDENTIFIER,
        V1HpQNormal._IDENTIFIER,
        V1HpLogNormal._IDENTIFIER,
        V1HpQLogNormal._IDENTIFIER,
    }:
        return None


def get_max(matrix):
    if matrix._IDENTIFIER == V1HpChoice._IDENTIFIER:
        if matrix.is_categorical:
            return None
        return max(to_numpy(matrix))

    if matrix._IDENTIFIER == V1HpPChoice._IDENTIFIER:
        return None

    if matrix._IDENTIFIER in {
        V1HpDateRange._IDENTIFIER,
        V1HpDateTimeRange._IDENTIFIER,
        V1HpRange._IDENTIFIER,
        V1HpLinSpace._IDENTIFIER,
        V1HpLogSpace._IDENTIFIER,
        V1HpGeomSpace._IDENTIFIER,
    }:
        return matrix.value.stop

    if matrix._IDENTIFIER == V1HpUniform._IDENTIFIER:
        return matrix.value.high

    if matrix._IDENTIFIER in {
        V1HpQUniform._IDENTIFIER,
        V1HpLogUniform._IDENTIFIER,
        V1HpQLogUniform._IDENTIFIER,
        V1HpNormal._IDENTIFIER,
        V1HpQNormal._IDENTIFIER,
        V1HpLogNormal._IDENTIFIER,
        V1HpQLogNormal._IDENTIFIER,
    }:
        return None


def to_numpy(matrix):
    if matrix._IDENTIFIER == V1HpChoice._IDENTIFIER:
        return matrix.value

    if matrix._IDENTIFIER == V1HpPChoice._IDENTIFIER:
        error_msg = (
            "Distribution should not call `to_numpy`, instead it should call `sample`."
        )
        if PYDANTIC_VERSION.startswith("2."):
            raise ValidationError.from_exception_data(
                title="Value Error",
                line_errors=[
                    {
                        "type": "value_error",
                        "loc": ("get_length",),
                        # Location tuple indicating where the error occurred
                        "msg": error_msg,
                        "input": matrix.__class__,  # The invalid input that caused the error
                        "ctx": {"error": error_msg},
                    }
                ],
            )
        else:
            raise ValidationError(  # TODO: Fix error message
                [
                    "Distribution should not call `to_numpy`, "
                    "instead it should call `sample`."
                ],
                matrix.__class__,
            )

    if matrix._IDENTIFIER == V1HpDateRange._IDENTIFIER:
        return np.arange(**matrix.value.to_dict()).astype(date)

    if matrix._IDENTIFIER == V1HpDateTimeRange._IDENTIFIER:
        return np.arange(**matrix.value.to_dict()).astype(datetime)

    if matrix._IDENTIFIER == V1HpRange._IDENTIFIER:
        return np.arange(**matrix.value.to_dict())

    if matrix._IDENTIFIER == V1HpLinSpace._IDENTIFIER:
        return np.linspace(**matrix.value.to_dict())

    if matrix._IDENTIFIER == V1HpLogSpace._IDENTIFIER:
        return np.logspace(**matrix.value.to_dict())

    if matrix._IDENTIFIER == V1HpGeomSpace._IDENTIFIER:
        return np.geomspace(**matrix.value.to_dict())

    if matrix._IDENTIFIER in {
        V1HpUniform._IDENTIFIER,
        V1HpQUniform._IDENTIFIER,
        V1HpLogUniform._IDENTIFIER,
        V1HpQLogUniform._IDENTIFIER,
        V1HpNormal._IDENTIFIER,
        V1HpQNormal._IDENTIFIER,
        V1HpLogNormal._IDENTIFIER,
        V1HpQLogNormal._IDENTIFIER,
    }:
        error_msg = (
            "Distribution should not call `to_numpy`, instead it should call `sample`."
        )
        if PYDANTIC_VERSION.startswith("2."):
            raise ValidationError.from_exception_data(
                title="Value Error",
                line_errors=[
                    {
                        "type": "value_error",
                        "loc": ("get_length",),
                        # Location tuple indicating where the error occurred
                        "msg": error_msg,
                        "input": matrix.__class__,  # The invalid input that caused the error
                        "ctx": {"error": error_msg},
                    }
                ],
            )
        else:
            raise ValidationError(
                [
                    "Distribution should not call `to_numpy`, "
                    "instead it should call `sample`."
                ],
                matrix.__class__,
            )


def _sample(matrix, size=None, rand_generator=None):
    size = None if size == 1 else size

    if matrix._IDENTIFIER == V1HpChoice._IDENTIFIER:
        return space_sample(
            value=to_numpy(matrix), size=size, rand_generator=rand_generator
        )
    if matrix._IDENTIFIER == V1HpPChoice._IDENTIFIER:
        return pchoice(values=matrix.value, size=size, rand_generator=rand_generator)

    if matrix._IDENTIFIER == V1HpDateRange._IDENTIFIER:
        return space_sample(
            value=to_numpy(matrix), size=size, rand_generator=rand_generator
        )

    if matrix._IDENTIFIER == V1HpDateTimeRange._IDENTIFIER:
        return space_sample(
            value=to_numpy(matrix), size=size, rand_generator=rand_generator
        )

    if matrix._IDENTIFIER == V1HpRange._IDENTIFIER:
        return space_sample(
            value=to_numpy(matrix), size=size, rand_generator=rand_generator
        )

    if matrix._IDENTIFIER == V1HpLinSpace._IDENTIFIER:
        return space_sample(
            value=to_numpy(matrix), size=size, rand_generator=rand_generator
        )

    if matrix._IDENTIFIER == V1HpLogSpace._IDENTIFIER:
        return space_sample(
            value=to_numpy(matrix), size=size, rand_generator=rand_generator
        )

    if matrix._IDENTIFIER == V1HpGeomSpace._IDENTIFIER:
        return space_sample(
            value=to_numpy(matrix), size=size, rand_generator=rand_generator
        )

    if matrix._IDENTIFIER == V1HpUniform._IDENTIFIER:
        return dist_sample(dist.uniform, matrix.value.to_dict(), size, rand_generator)

    if matrix._IDENTIFIER == V1HpQUniform._IDENTIFIER:
        return dist_sample(dist.quniform, matrix.value.to_dict(), size, rand_generator)

    if matrix._IDENTIFIER == V1HpLogUniform._IDENTIFIER:
        return dist_sample(
            dist.loguniform, matrix.value.to_dict(), size, rand_generator
        )

    if matrix._IDENTIFIER == V1HpQLogUniform._IDENTIFIER:
        return dist_sample(
            dist.qloguniform, matrix.value.to_dict(), size, rand_generator
        )

    if matrix._IDENTIFIER == V1HpNormal._IDENTIFIER:
        return dist_sample(dist.normal, matrix.value.to_dict(), size, rand_generator)

    if matrix._IDENTIFIER == V1HpQNormal._IDENTIFIER:
        return dist_sample(dist.qnormal, matrix.value.to_dict(), size, rand_generator)

    if matrix._IDENTIFIER == V1HpLogNormal._IDENTIFIER:
        return dist_sample(dist.lognormal, matrix.value.to_dict(), size, rand_generator)

    if matrix._IDENTIFIER == V1HpQLogNormal._IDENTIFIER:
        return dist_sample(
            dist.qlognormal, matrix.value.to_dict(), size, rand_generator
        )


def sample(matrix, size=None, rand_generator=None):
    try:
        return _sample(matrix, size=size, rand_generator=rand_generator)
    except Exception as e:
        error_msg = (
            "Could not sample from matrix value: {} for kind: {} with size: {}".format(
                matrix.value, matrix._IDENTIFIER, size
            )
        )
        if PYDANTIC_VERSION.startswith("2."):
            raise ValidationError.from_exception_data(
                title="Value Error",
                line_errors=[
                    {
                        "type": "value_error",
                        "loc": ("get_length",),
                        # Location tuple indicating where the error occurred
                        "msg": error_msg,
                        "input": matrix.__class__,  # The invalid input that caused the error
                        "ctx": {"error": error_msg},
                    }
                ],
            ) from e
        else:
            raise ValidationError(
                [error_msg],
                matrix.__class__,
            ) from e



import hyperopt

from hypertune.matrix.utils import to_numpy
from polyaxon.schemas import (
    V1HpChoice,
    V1HpGeomSpace,
    V1HpLinSpace,
    V1HpLogNormal,
    V1HpLogSpace,
    V1HpLogUniform,
    V1HpNormal,
    V1HpPChoice,
    V1HpQLogNormal,
    V1HpQLogUniform,
    V1HpQNormal,
    V1HpQUniform,
    V1HpRange,
    V1HpUniform,
)


def to_hyperopt(name, matrix):
    if matrix._IDENTIFIER in {
        V1HpChoice._IDENTIFIER,
        V1HpRange._IDENTIFIER,
        V1HpLinSpace._IDENTIFIER,
        V1HpLogSpace._IDENTIFIER,
        V1HpGeomSpace._IDENTIFIER,
    }:
        return hyperopt.hp.choice(name, to_numpy(matrix))

    if matrix._IDENTIFIER == V1HpPChoice._IDENTIFIER:
        raise ValueError("{} is not supported by Hyperopt.".format(matrix._IDENTIFIER))

    if matrix._IDENTIFIER == V1HpUniform._IDENTIFIER:
        return hyperopt.hp.uniform(name, matrix.value.low, matrix.value.high)

    if matrix._IDENTIFIER == V1HpQUniform._IDENTIFIER:
        return hyperopt.hp.quniform(
            name,
            matrix.value.low,
            matrix.value.high,
            matrix.value.q,
        )

    if matrix._IDENTIFIER == V1HpLogUniform._IDENTIFIER:
        return hyperopt.hp.loguniform(name, matrix.value.low, matrix.value.high)

    if matrix._IDENTIFIER == V1HpQLogUniform._IDENTIFIER:
        return hyperopt.hp.qloguniform(
            name,
            matrix.value.low,
            matrix.value.high,
            matrix.value.q,
        )

    if matrix._IDENTIFIER == V1HpNormal._IDENTIFIER:
        return hyperopt.hp.normal(name, matrix.value.loc, matrix.value.scale)

    if matrix._IDENTIFIER == V1HpQNormal._IDENTIFIER:
        return hyperopt.hp.qnormal(
            name,
            matrix.value.loc,
            matrix.value.scale,
            matrix.value.q,
        )

    if matrix._IDENTIFIER == V1HpLogNormal._IDENTIFIER:
        return hyperopt.hp.lognormal(name, matrix.value.loc, matrix.value.scale)

    if matrix._IDENTIFIER == V1HpQLogNormal._IDENTIFIER:
        return hyperopt.hp.qlognormal(
            name,
            matrix.value.loc,
            matrix.value.scale,
            matrix.value.q,
        )



NAME = "hypertune"
VERSION = "1.2.1-rc1"
DESC = "A library for performing hyperparameter optimization with Polyaxon."
URL = "https://github.com/polyaxon/hypertune"
AUTHOR = "Polyaxon, Inc."
EMAIL = "contact@polyaxon.com"
LICENSE = "Apache 2.0"



import itertools

from typing import Dict, List

from hypertune.matrix.utils import to_numpy
from hypertune.search_managers.base import BaseManager
from polyaxon.schemas import V1GridSearch


class GridSearchManager(BaseManager):
    """Grid search strategy manager for hyperparameter optimization."""

    CONFIG = V1GridSearch

    def get_suggestions(self, params: List[Dict] = None) -> List[Dict]:
        suggestions = []
        keys = list(self.config.params.keys())
        values = [to_numpy(v) for v in self.config.params.values()]
        for v in itertools.product(*values):
            suggestions.append(dict(zip(keys, v)))

        if self.config.num_runs:
            return suggestions[: self.config.num_runs]
        return suggestions



import copy

from functools import reduce
from operator import mul
from typing import Dict, List, Optional

from hypertune.matrix.utils import get_length, sample
from hypertune.search_managers.base import BaseManager
from hypertune.search_managers.spec import SuggestionSpec
from hypertune.search_managers.utils import get_random_generator
from polyaxon.schemas import V1RandomSearch


class RandomSearchManager(BaseManager):
    """Random search strategy manager for hyperparameter optimization."""

    CONFIG = V1RandomSearch

    def get_suggestions(self, params: Optional[Dict] = None) -> List[Dict]:
        if not self.config.num_runs:
            raise ValueError("This search strategy requires `num_runs`.")
        suggestions = []
        params = params or {}
        rand_generator = get_random_generator(seed=self.config.seed)
        # Validate number of suggestions and total space
        all_discrete = True
        for v in self.config.params.values():
            if v.is_continuous:
                all_discrete = False
                break
        num_runs = self.config.num_runs
        if all_discrete:
            space = reduce(mul, [get_length(v) for v in self.config.params.values()])
            num_runs = self.config.num_runs if self.config.num_runs <= space else space

        while num_runs > 0:
            suggestion_params = copy.deepcopy(params)
            suggestion_params.update(
                {
                    k: sample(v, rand_generator=rand_generator)
                    for k, v in self.config.params.items()
                }
            )
            suggestion = SuggestionSpec(params=suggestion_params)
            if suggestion not in suggestions:
                suggestions.append(suggestion)
                num_runs -= 1
        return [suggestion.params for suggestion in suggestions]



import copy

from typing import Dict, List, Optional

from hypertune.search_managers.base import BaseManager
from polyaxon.schemas import V1Mapping


class MappingManager(BaseManager):
    """Mapping strategy manager for running parallel operations."""

    CONFIG = V1Mapping

    def get_suggestions(self, params: Optional[Dict] = None) -> List[Dict]:
        suggestions = []
        params = params or {}
        for v in self.config.values:
            suggestion_params = copy.deepcopy(params)
            suggestion_params.update(copy.deepcopy(v))
            suggestions.append(suggestion_params)
        return suggestions



from typing import Dict, List


class BaseManager:
    CONFIG = None

    def __init__(self, config):
        if not isinstance(config, self.CONFIG):
            raise ValueError(
                "The current search manager `{}` is not compatible "
                "with the search kind `{}` defined in the config.".format(
                    self.CONFIG._IDENTIFIER, config._IDENTIFIER
                )
            )
        self.config = config

    def get_suggestions(self, **kwargs) -> List[Dict]:
        raise NotImplementedError



import numpy as np


def uniform(low, high, size=None, rand_generator=None):
    rand_generator = rand_generator or np.random
    return rand_generator.uniform(low=low, high=high, size=size)


def quniform(low, high, q, size=None, rand_generator=None):
    value = uniform(low=low, high=high, size=size, rand_generator=rand_generator)
    return np.round(value // q) * q


def loguniform(low, high, size=None, rand_generator=None):
    value = uniform(low=low, high=high, size=size, rand_generator=rand_generator)
    return np.exp(value)


def qloguniform(low, high, q, size=None, rand_generator=None):
    value = loguniform(low=low, high=high, size=size, rand_generator=rand_generator)
    return np.round(value // q) * q


def normal(loc, scale, size=None, rand_generator=None):
    rand_generator = rand_generator or np.random
    return rand_generator.normal(loc=loc, scale=scale, size=size)


def qnormal(loc, scale, q, size=None, rand_generator=None):
    draw = normal(loc=loc, scale=scale, size=size, rand_generator=rand_generator)
    return np.round(draw // q) * q


def lognormal(loc, scale, size=None, rand_generator=None):
    rand_generator = rand_generator or np.random
    return rand_generator.lognormal(mean=loc, sigma=scale, size=size)


def qlognormal(loc, scale, q, size=None, rand_generator=None):
    draw = lognormal(loc=loc, scale=scale, size=size, rand_generator=rand_generator)
    return np.round(draw // q) * q
